@@ -8,23 +8,22 @@ time. If functions defined in friendly_traceback.__init__.py do not meet your ne
 please file an issue.
 """
 import inspect
-import os
 import re
 import traceback
 
 from itertools import dropwhile
 
+from . import debug_helper
 from . import info_generic
 from . import info_specific
 from . import info_variables
-from . import debug_helper
-from . import source_cache
 
+from .frame_info import FrameInfo
 from .ft_gettext import current_lang
 
 from .path_info import is_excluded_file, EXCLUDED_FILE_PATH, path_utils
 from .runtime_errors import name_error
-from .source_cache import cache, highlight_source
+from .source_cache import cache
 from .syntax_errors import analyze_syntax
 from .syntax_errors import indentation_error
 from .syntax_errors import source_info
@@ -108,29 +107,44 @@ class TracebackData:
             self.statement.bad_line = remove_space(self.statement.bad_line)
         else:
             self.statement = None
-            self.locate_error(tb)
+            self.locate_error()
 
     def get_records(self, tb):
         """Get the traceback frame history, excluding those originating
         from our own code that are included either at the beginning or
         at the end of the traceback.
         """
-        records = inspect.getinnerframes(tb, cache.context)
-        records = list(
-            dropwhile(lambda record: is_excluded_file(record.filename), records)
-        )
-        records.reverse()
-        records = list(
-            dropwhile(lambda record: is_excluded_file(record.filename), records)
-        )
-        records.reverse()
-        if records or issubclass(self.exception_type, SyntaxError):
-            return records
-        # If all the records are removed, it means that all the error
+        try:
+            all_records = list(FrameInfo.stack_data(tb, collapse_repeated_frames=False))
+            records = list(
+                dropwhile(lambda record: is_excluded_file(record.filename), all_records)
+            )
+            records.reverse()
+            records = list(
+                dropwhile(lambda record: is_excluded_file(record.filename), records)
+            )
+            records.reverse()
+            if records or issubclass(self.exception_type, (SyntaxError, MemoryError)):
+                return records
+        except AssertionError:  # from stack_data
+            # problems may arise when SyntaxErrors are raise
+            # from a normal console like the one used in Mu.
+            all_records = inspect.getinnerframes(tb, cache.context)
+            records = list(
+                dropwhile(lambda record: is_excluded_file(record.filename), all_records)
+            )
+            records.reverse()
+            records = list(
+                dropwhile(lambda record: is_excluded_file(record.filename), records)
+            )
+            records.reverse()
+            if records or issubclass(self.exception_type, (SyntaxError, MemoryError)):
+                return records
+        # If all the records are removed, it likely means that all the error
         # is in our own code - or that of the user who chose to exclude
         # some files. If so, we make sure to have something to analyze
         # and help identify the problem.
-        return inspect.getinnerframes(tb, cache.context)  # pragma: no cover
+        return all_records  # pragma: no cover
 
     def get_source_info(self):
         """Retrieves the file name and the line of code where the exception
@@ -161,19 +175,31 @@ class TracebackData:
             return
 
         if self.records:
-            self.exception_frame, self.filename, linenumber, _, _, _ = self.records[-1]
-            _, line = cache.get_formatted_partial_source(self.filename, linenumber)
+            record = self.records[-1]
+            self.exception_frame = record.frame
+            self.filename = record.filename
+            line = record.problem_line()
             self.bad_line = line.rstrip()
+            # protecting against https://github.com/alexmojaki/stack_data/issues/13
+            if not self.bad_line:
+                try:
+                    lines = cache.get_source_lines(record.filename)
+                    self.bad_line = lines[record.lineno - 1]
+                except Exception:  # noqa
+                    debug_helper.log("Could not get bad_line")
+
             if len(self.records) > 1:
-                self.program_stopped_frame, filename, linenumber, *_rest = self.records[
-                    0
-                ]
-                _, line = cache.get_formatted_partial_source(filename, linenumber)
+                record = self.records[0]
+                line = record.problem_line()
+                self.program_stopped_frame = record.frame
                 self.program_stopped_bad_line = line.rstrip()
             else:
                 self.program_stopped_bad_line = self.bad_line
                 self.program_stopped_frame = self.exception_frame
             return
+        elif issubclass(self.exception_type, MemoryError):
+            self.bad_line = "<not available>"
+            return
 
         # We should never reach this stage.
         def _log_error():  # pragma: no cover
@@ -185,35 +211,24 @@ class TracebackData:
 
         _log_error()  # pragma: no cover
 
-    def locate_error(self, tb):
+    def locate_error(self):
         """Attempts to narrow down the location of the error so that,
         if possible, the problem code is highlighted with ^^^^."""
         if not self.records:  # pragma: no cover
+            if issubclass(self.exception_type, MemoryError):
+                return
             debug_helper.log("No records in locate_error().")
             return
 
-        if self.program_stopped_frame is not None:
-            exc_tb = self.find_tb_frame(tb, self.program_stopped_frame)
-            if exc_tb is not None:
-                _, self.program_stopped_node_range, _ = self.find_node(
-                    exc_tb, self.program_stopped_bad_line
-                )
-        if self.exception_name == "NameError":
-            # `executing` cannot give us the node location in this case
-            return self.locate_name_error()
-
-        tb = self.find_tb_frame(tb, self.exception_frame)
-        if tb is None:
-            debug_helper.log("Exception frame could not be found.")  # pragma: no cover
-            return  # pragma: no cover
-
-        self.node, self.node_range, self.node_text = self.find_node(tb, self.bad_line)
-        if self.node_text.strip():
-            # Replacing the line that caused the exception by the text
-            # of the 'node' facilitates the process of identifying the cause.
-            # However, in a few cases, we do need to keep the entire original line.
-            self.original_bad_line = self.bad_line
-            self.bad_line = self.node_text
+        node_info = self.records[-1].node_info  # noqa
+        if node_info:
+            self.node, _, self.node_text = node_info
+            if self.node_text.strip():
+                # Replacing the line that caused the exception by the text
+                # of the 'node' facilitates the process of identifying the cause.
+                # However, in a few cases, we do need to keep the entire original line.
+                self.original_bad_line = self.bad_line
+                self.bad_line = self.node_text
 
     @staticmethod
     def find_tb_frame(tb, frame):
@@ -541,6 +556,8 @@ class FriendlyTraceback:
 
         records = self.tb_data.records
         if not records:  # pragma: no cover
+            if issubclass(self.tb_data.exception_type, MemoryError):
+                return
             debug_helper.log("No record in assign_location().")
             return
 
@@ -556,19 +573,15 @@ class FriendlyTraceback:
         * exception_raised_source
         * exception_raised_variables
         """
+        from .config import session
+
         _ = current_lang.translate
 
-        frame, filename, linenumber, _func, lines, index = record
-        if (
-            lines == ["\n"] and source_cache.idle_get_lines is not None
-        ):  # pragma: no cover
-            # skipcq: PYL-E1102
-            lines = source_cache.idle_get_lines(filename, linenumber - 1)
-
-        partial_source = get_partial_source(
-            filename, linenumber, lines, index, self.tb_data.node_range
+        # partial_source = get_partial_source(record, text_range=self.tb_data.node_range)
+        partial_source = record.partial_source_with_node_range
+        filename = path_utils.shorten_path(
+            record.filename, frame=self.tb_data.exception_frame
         )
-        filename = path_utils.shorten_path(filename)
 
         unavailable = filename in ["<unknown>", "<string>"]
         if unavailable:
@@ -576,9 +589,14 @@ class FriendlyTraceback:
                 "{filename} is not a regular Python file whose contents can be analyzed.\n"
             ).format(filename=filename)
 
-        self.info["exception_raised_header"] = _(
-            "Exception raised on line {linenumber} of file {filename}.\n"
-        ).format(linenumber=linenumber, filename=filename)
+        if session.ipython_prompt:
+            self.info["exception_raised_header"] = _(
+                "Exception raised on line {linenumber} of code block {filename}.\n"
+            ).format(linenumber=record.lineno, filename=filename)
+        else:
+            self.info["exception_raised_header"] = _(
+                "Exception raised on line {linenumber} of file {filename}.\n"
+            ).format(linenumber=record.lineno, filename=filename)
 
         if unavailable:
             return
@@ -590,7 +608,7 @@ class FriendlyTraceback:
         else:
             line = partial_source["line"]
 
-        var_info = info_variables.get_var_info(line, frame)
+        var_info = info_variables.get_var_info(line, record.frame)
         if var_info:
             self.info["exception_raised_variables"] = var_info
 
@@ -601,23 +619,28 @@ class FriendlyTraceback:
         * exception_raised_source
         * last_call_variables
         """
+        from .config import session
+
         _ = current_lang.translate
-
-        frame, filename, linenumber, _func, lines, index = record
-        if lines == ["\n"] and source_cache.idle_get_lines is not None:
-            # skipcq: PYL-E1102
-            lines = source_cache.idle_get_lines(filename, linenumber - 1)
-        partial_source = get_partial_source(
-            filename, linenumber, lines, index, self.tb_data.program_stopped_node_range
+        partial_source = record.partial_source_with_node_range
+        # partial_source = get_partial_source(
+        #     record, text_range=self.tb_data.program_stopped_node_range
+        # )
+        filename = path_utils.shorten_path(
+            record.filename, frame=self.tb_data.exception_frame
         )
-        filename = path_utils.shorten_path(filename)
 
-        self.info["last_call_header"] = _(
-            "Execution stopped on line {linenumber} of file {filename}.\n"
-        ).format(linenumber=linenumber, filename=filename)
+        if session.ipython_prompt:
+            self.info["last_call_header"] = _(
+                "Execution stopped on line {linenumber} of code block {filename}.\n"
+            ).format(linenumber=record.lineno, filename=filename)
+        else:
+            self.info["last_call_header"] = _(
+                "Execution stopped on line {linenumber} of file {filename}.\n"
+            ).format(linenumber=record.lineno, filename=filename)
         self.info["last_call_source"] = partial_source["source"]
 
-        var_info = info_variables.get_var_info(partial_source["line"], frame)
+        var_info = info_variables.get_var_info(partial_source["line"], record.frame)
         if var_info:
             self.info["last_call_variables"] = var_info
 
@@ -627,6 +650,8 @@ class FriendlyTraceback:
         * parsing_error
         * parsing_source_error
         """
+        from .config import session
+
         _ = current_lang.translate
         value = self.tb_data.value
         filepath = value.filename
@@ -641,16 +666,25 @@ class FriendlyTraceback:
         statement.format_statement()
         partial_source = statement.formatted_partial_source
 
-        if "-->" in partial_source:
-            self.info["parsing_error"] = _(
-                "Python could not understand the code in the file\n"
-                "'{filename}'\n"
-                "beyond the location indicated by ^.\n"
-            ).format(filename=path_utils.shorten_path(filepath))
+        short_filename = path_utils.shorten_path(
+            filepath, frame=self.tb_data.exception_frame
+        )
+
+        if session.ipython_prompt:
+            could_not_understand = _(
+                "Python could not understand the code in the code block {filename}\n"
+            ).format(filename=short_filename)
+        else:
+            could_not_understand = _(
+                "Python could not understand the code in the file\n" "'{filename}'\n"
+            ).format(filename=short_filename)
+
+        if "-->" in partial_source and "^" in partial_source:
+            self.info["parsing_error"] = could_not_understand + _(
+                "at the location indicated by ^.\n"
+            ).format(filename=short_filename)
         elif filepath:  # could be None
-            self.info["parsing_error"] = _(
-                "Python could not understand the code in the file\n" "'{filename}'.\n"
-            ).format(filename=path_utils.shorten_path(filepath))
+            self.info["parsing_error"] = could_not_understand + ".\n"
 
         self.info["parsing_error_source"] = f"{partial_source}\n"
 
@@ -689,7 +723,6 @@ class FriendlyTraceback:
             return
 
         python_tb = [line.rstrip() for line in self.tb_data.formatted_tb]
-
         tb = self.create_traceback()
         shortened_tb = self.shorten(tb)
 
@@ -744,16 +777,27 @@ class FriendlyTraceback:
         """Shortens a traceback (as list of lines)
         by removing lines if it exceeds a certain length
         and by using short synonyms for some common directories."""
+        from .config import session
+
         _ = current_lang.translate
         shortened_tb = tb[0:2] + self.suppressed + tb[-5:] if len(tb) > 12 else tb[:]
-        pattern = re.compile(r'^  File "(.*)", ')
+        pattern = re.compile(r'^  File "(.*)", ')  # noqa
         temp = []
         for line in shortened_tb:
             match = re.search(pattern, line)
             if match:
-                line = line.replace(
-                    match.group(1), path_utils.shorten_path(match.group(1))
+                filename = match.group(1)
+                short_filename = path_utils.shorten_path(
+                    filename, frame=self.tb_data.exception_frame
                 )
+                line = line.replace(filename, short_filename)
+                if (
+                    session.ipython_prompt
+                    and short_filename[0] == "["
+                    and short_filename[-1] == "]"
+                ):
+                    line = line.replace("  File", "  Code block")
+                    line = line.replace('"[', "[").replace(']"', "]")
             temp.append(line)
         return temp
 
@@ -802,10 +846,12 @@ class FriendlyTraceback:
         """
         result = []
         for record in self.tb_data.records:
-            frame, filename, linenumber, _func, lines, index = record
-            partial_source = get_partial_source(filename, linenumber, lines, index)
+            partial_source = record.partial_source  # noqa
+            # partial_source = get_partial_source(record)
             result.append(
-                '  File "{}", line {}, in {}'.format(filename, linenumber, _func)
+                '  File "{}", line {}, in {}'.format(
+                    record.filename, record.lineno, record.code.co_name
+                )
             )
             bad_line = partial_source["line"]
             if bad_line is not None:
@@ -853,47 +899,67 @@ class FriendlyTraceback:
         return result
 
 
-def get_partial_source(filename, linenumber, lines, index, text_range=None):
-    """Gets the part of the source where an exception occurred,
-    formatted in a pre-determined way, as well as the content
-    of the specific line where the exception occurred.
-    """
-    _ = current_lang.translate
-
-    file_not_found = _("Problem: source of `{filename}` is not available\n").format(
-        filename=filename
-    )
-    if filename in cache.cache:
-        source, line = cache.get_formatted_partial_source(
-            filename, linenumber, text_range=text_range
-        )
-    elif filename and os.path.abspath(filename):
-        source, line = highlight_source(linenumber, index, lines, text_range=text_range)
-        if not source:  # pragma: no cover
-            line = ""
-            if filename == "<stdin>":
-                source = "\n"  # Using a normal Python REPL - source unavailable.
-                # An appropriate error message will have been given via
-                # cannot_analyze_stdin
-            else:
-                source = file_not_found
-                debug_helper.log("Problem in get_partial_source().")
-                debug_helper.log(file_not_found)
-    elif not filename:  # pragma: no cover
-        source = file_not_found
-        line = ""
-        debug_helper.log("Problem in get_partial_source().")
-        debug_helper.log(file_not_found)
-    else:  # pragma: no cover
-        source = line = ""
-        debug_helper.log("Problem in get_partial_source().")
-        debug_helper.log("Should not have reached this option")
-        debug_helper.log_error()
-
-    if not source.endswith("\n"):
-        source += "\n"
-
-    return {"source": source, "line": line}
+# def get_partial_source(record, text_range=None):
+#     """Gets the part of the source where an exception occurred,
+#     formatted in a pre-determined way, as well as the content
+#     of the specific line where the exception occurred.
+#     """
+#     _ = current_lang.translate
+#     file_not_found = _("Problem: source of `{filename}` is not available\n").format(
+#         filename=record.filename
+#     )
+#     source = line = ""
+#
+#     lines = record.lines
+#     if not lines and record.filename:
+#         # protecting against https://github.com/alexmojaki/stack_data/issues/13
+#         try:
+#             lineno = record.lineno
+#             s_lines = cache.get_source_lines(record.filename)
+#
+#             class FakeLineObject:
+#                 def __init__(self, line, linenumber, lineno):
+#                     self.text = line
+#                     self.lineno = linenumber + 1
+#                     self.is_current = self.lineno == lineno
+#
+#             lines = []
+#             linenumber = lineno - 2
+#             for line in s_lines[linenumber : lineno + 1]:
+#                 lines.append(FakeLineObject(line, linenumber, lineno))
+#                 linenumber += 1
+#         except Exception:  # noqa
+#             pass
+#
+#     if lines:
+#         source, line = highlight_source(lines, text_range=text_range)
+#     elif record.filename and os.path.abspath(record.filename):
+#         if record.filename == "<stdin>":
+#             pass
+#             # Using a normal Python REPL - source unavailable.
+#             # An appropriate error message will have been given via
+#             # cannot_analyze_stdin
+#         else:
+#             source = file_not_found
+#             debug_helper.log("Problem in get_partial_source().")
+#             debug_helper.log(file_not_found)
+#             debug_helper.log(str(record))
+#     elif not record.filename:  # pragma: no cover
+#         source = file_not_found
+#         line = ""
+#         debug_helper.log("Problem in get_partial_source(): no file name")
+#         debug_helper.log(file_not_found)
+#         debug_helper.log(str(record))
+#     else:  # pragma: no cover
+#         source = line = ""
+#         debug_helper.log("Problem in get_partial_source().")
+#         debug_helper.log("Should not have reached this option")
+#         debug_helper.log_error()
+#
+#     if not source.endswith("\n"):
+#         source += "\n"
+#
+#     return {"source": source, "line": line}
 
 
 def cannot_analyze_stdin():  # pragma: no cover
@@ -906,4 +972,5 @@ def cannot_analyze_stdin():  # pragma: no cover
         "Unfortunately, no additional information is available:\n"
         "the content of file '<stdin>' is not accessible.\n"
         "Are you using a regular Python console instead of a Friendly-console?\n"
+        "Perhaps try: `start_console(local_vars=locals())`.\n"
     )
